/* jabber application
 *
 * author: Ben Fuhrmannek <bef@erlangen.ccc.de>
 * version: 0.1-devel-20051202
 * license: GPL
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <sys/signal.h>

#include <loudmouth/loudmouth.h>

#include "asterisk.h"

ASTERISK_FILE_VERSION(__FILE__, "$Revision: 1.20 $")

#include "asterisk/file.h"
#include "asterisk/logger.h"
#include "asterisk/channel.h"
#include "asterisk/pbx.h"
#include "asterisk/module.h"
#include "asterisk/lock.h"
#include "asterisk/app.h"
#include "asterisk/cli.h"


static const char desc[] = "jabber application";
static const char config[] = "jabber.conf";

/* debugging */
#define JD_JABBER (1 << 0)
#define JD_ASTERISK (1 << 1)
#define jdebug(__level, __fmt, __arg...) if (jabberdebug & (__level)) ast_log(LOG_NOTICE, __fmt, ## __arg)
static int jabberdebug = 0xffff;

/* foo */
STANDARD_LOCAL_USER;
LOCAL_USER_DECL;


/****************************************************************************/
/*** STRUCTS & PROTOTYPES ***/
/****************************************************************************/
/* jabber init */
int jabber_load(void);
int jabber_unload(void);
int read_config(void);

/****************************************************************************/
/* jabber accounts */
struct jabber_account {
	/* configurable options */
	char user[AST_MAX_EXTENSION];
	char host[AST_MAX_EXTENSION];
	char port[AST_MAX_EXTENSION];
	char pass[AST_MAX_EXTENSION];
	char resource[AST_MAX_EXTENSION];
	int use_ssl;
	char ssl_fingerprint[AST_MAX_EXTENSION];

	/* status flags & auto-config options*/
	char config_section[AST_MAX_EXTENSION]; /* unique key!! */

	/* jabber client */
	pthread_t client_thread;
	void *client_data;
	char partner_jid[AST_MAX_EXTENSION];
	char new_message[AST_MAX_EXTENSION];

	/* list */
	struct jabber_account *next;
};
static struct jabber_account *jabber_account_list = NULL;

static struct jabber_account* create_new_jabber_account(void);
static int clear_jabber_account_list(void);
static struct jabber_account *find_jabber_account(char *config_section);

/****************************************************************************/
/* jabber client */

static void *do_jabber_client(void *data);
struct jabber_client_data {
	GMainLoop *main_loop;
	LmConnection *connection;
};
static int jabber_send_message(const char *dialstring, const char *message);
static int jabber_client_send_message(LmConnection *connection, const char *recipient_jid, const char *message);
static int jid_compare(const char *jid, const char *jid_and_resource);
static LmHandlerResult jabber_message_handler(LmMessageHandler *handler, LmConnection *connection, LmMessage *m, gpointer user_data);
static int split_dialstring(char *tmp_dialstring, char **config_section, char **jid);



/****************************************************************************/
/* JABBER CLIENT */
/****************************************************************************/

/* compare a jid such as "user@host" with a jid+resource such as "user@host/app" */
static int jid_compare(const char *jid, const char *jid_and_resource)
{
	char tmp[256];
	char *res;
	ast_copy_string(tmp, jid_and_resource, sizeof(tmp));
	res = strchr(tmp, '/');
	if (res)
		*res = '\0';
	return strncmp(jid, tmp, sizeof(tmp));
}

static LmHandlerResult jabber_message_handler(LmMessageHandler *handler, LmConnection *connection, LmMessage *m, gpointer user_data)
{
	const char *from;
	const char *message;
	struct jabber_account *jac, *jac2;

	/* get sender and message */
	from = lm_message_node_get_attribute(m->node, "from"); /*!< foo@bar/app */
	message = lm_message_node_get_child(m->node, "body")->value; /*!< <message><body>...</body></message> */
	
	jdebug(JD_JABBER, "incoming message from %s: %s\n", from, message);

	/* find first matching account */
	jac = jabber_account_list;
	jac2 = NULL;
	while (jac) {
		if (!jid_compare(jac->partner_jid, from))
		{
			jac2 = jac;
			break;
		}
		else
			jac = jac->next;
	}

	if (jac2)
	{
		if (jac2->new_message[0])
			jdebug(JD_JABBER, "the previous message will be lost\n");
		/* buffer last message only -- messages may get lost this way*/
		ast_copy_string(jac2->new_message, message, sizeof(jac2->new_message));
	}
	else
	{
		jdebug(JD_JABBER, "ignoring message\n");
	}

	return LM_HANDLER_RESULT_REMOVE_MESSAGE;
}

static int split_dialstring(char *tmp_dialstring, char **config_section, char **jid)
{
	*jid = strchr(tmp_dialstring, '/');
	if (!*jid) {
		return -1;
	}
	*config_section = tmp_dialstring;
	**jid = '\0';
	(*jid)++;

	return 0;
}
/* send jabber message | dialstring is "config_section/user@host" as in Dial(JABBER/...) app */
static int jabber_send_message(const char *dialstring, const char *message)
{
	struct jabber_client_data *jcd;
	char *recipient_jid;
	char *config_section;
	char tmp[256]; /* may be too small? */
	struct jabber_account *jac;

	/* extract jid & config_section from dialstring */
	ast_copy_string(tmp, dialstring, sizeof(tmp));
	if (split_dialstring(tmp, &config_section, &recipient_jid)) {
		ast_log(LOG_ERROR, "incorrect format for Dial(...)\n");
		return -1;
	}
	
	/* find matching account */
	jac = find_jabber_account(config_section);
	if (!jac) {
		ast_log(LOG_ERROR, "unknown config section [%s]\n", config_section);
		return -1;
	}
	jdebug(JD_JABBER, "sending msg to %s using %s\n", recipient_jid, jac->config_section);

	/* get connection */
	if (!jac->client_data) {
		ast_log(LOG_ERROR, "client appears to be uninitialized [%s]\n", config_section);
		return -1;
	}
	jcd = jac->client_data;
	if (!jcd->connection) {
		ast_log(LOG_ERROR, "client is offline [%s]\n", config_section);
		return -1;
	}
	
	/* send message */
	return jabber_client_send_message(jcd->connection, recipient_jid, message);
}

/* send message without further checks */
static int jabber_client_send_message(LmConnection *connection, const char *recipient_jid, const char *message)
{
	LmMessage *m;
	gboolean   result;
	GError *error = NULL;
	
	/* assemble new message */
	m = lm_message_new (recipient_jid, LM_MESSAGE_TYPE_MESSAGE);
	lm_message_node_add_child (m->node, "body", message);
	
	/* send message */
	result = lm_connection_send (connection, m, &error);
	lm_message_unref (m);
	
	if (!result) {
		ast_log(LOG_ERROR, "lm_connection_send failed\n");
		return -1;
	}
	
	return 0;
}

/* send presence specified by sub_type */
static int jabber_client_send_presence(LmConnection *connection, LmMessageSubType sub_type)
{
	LmMessage *m;
	gboolean   result;
	GError *error = NULL;

	/* assemble new message */
	m = lm_message_new_with_sub_type(NULL, LM_MESSAGE_TYPE_PRESENCE, sub_type);
	
	/* send message */
	result = lm_connection_send (connection, m, &error);
	lm_message_unref (m);
	
	if (!result) {
		ast_log(LOG_ERROR, "lm_connection_send failed\n");
		return -1;
	}
	
	return 0;
}

/* authentication done (callback), let's send presence */
static void jabber_client_connection_auth_cb(LmConnection *connection, gboolean success, struct jabber_account *jac)
{
	if (!success) {
		ast_log(LOG_ERROR, "Connection failed");
		return;
	}

	jdebug(JD_JABBER, "connected to %s [%s]\n", jac->host, jac->config_section);

	/* jabber_client_send_message(connection, "admin@arbeitszimmer", "-= connected =-"); */
	jabber_client_send_presence(connection, LM_MESSAGE_SUB_TYPE_AVAILABLE);
}

/* connection is open (callback), we can authenticate now. */
static void jabber_client_connection_open_result_cb(LmConnection *connection, gboolean success, struct jabber_account *jac)
{
	GError *error = NULL;
	
	if (!success) {
		ast_log(LOG_ERROR, "Connection failed\n");
		return;
	}

	if (!lm_connection_authenticate (connection, jac->user,
					 jac->pass, jac->resource,
					 (LmResultFunction) jabber_client_connection_auth_cb,
					 jac,
					 NULL,
					 &error)) {
		ast_log(LOG_ERROR, "lm_connection_authenticate failed\n");
		return;
	}
}

/* handle ssl errors */
static LmSSLResponse jabber_client_ssl_error_handler(LmSSL *ssl, LmSSLStatus status, gpointer user_data)
{
	struct jabber_account *jac = user_data;
	
	jdebug(JD_JABBER, "jabber: ssl error [status=%d]\n", status);

	if (strcmp(jac->ssl_fingerprint, "")) { /* we have a fingerprint */
		
		if (status == LM_SSL_STATUS_NO_CERT_FOUND) {
			jdebug(JD_JABBER, "jabber: no cert found\n");
			return LM_SSL_RESPONSE_STOP;
		}
	
		if (status == LM_SSL_STATUS_CERT_FINGERPRINT_MISMATCH) {
			jdebug(JD_JABBER, "jabber: unmatched fingerprints [%s != %s]\n", lm_ssl_get_fingerprint(ssl), jac->ssl_fingerprint);
			return LM_SSL_RESPONSE_STOP;
		}
	}

	return LM_SSL_RESPONSE_CONTINUE;
}

static void *do_jabber_client(void *data)
{
	struct jabber_account *jac = data;
	struct jabber_client_data *jcd;
	LmSSL *ssl;
	char *fingerprint = NULL;
	LmMessageHandler *handler;
	GMainContext *context;
	GError       *error = NULL;

	jcd = malloc(sizeof(struct jabber_client_data));
	jac->client_data = jcd;

	/* new connection */
	context = g_main_context_new ();
	jcd->connection = lm_connection_new_with_context (jac->host, context);

	/* port */
	lm_connection_set_port(jcd->connection, atoi(jac->port));
			
	/* ssl */
	if (jac->use_ssl) {
		
		fingerprint = NULL;
		if (!strcmp(jac->ssl_fingerprint, ""))
				fingerprint = jac->ssl_fingerprint;
		
		ssl = lm_ssl_new(fingerprint, jabber_client_ssl_error_handler, jac, NULL);
		lm_connection_set_ssl(jcd->connection, ssl);
		
		lm_ssl_unref(ssl);
	}
	
	/* message handler */
	handler = lm_message_handler_new (jabber_message_handler, NULL, NULL);
	lm_connection_register_message_handler(jcd->connection, handler, LM_MESSAGE_TYPE_MESSAGE, LM_HANDLER_PRIORITY_NORMAL);
	lm_message_handler_unref(handler);
	
	/* open connection */
	if (!lm_connection_open (jcd->connection, (LmResultFunction) jabber_client_connection_open_result_cb, jac, NULL, &error)) {
		ast_log(LOG_ERROR, "lm_connection_open failed");
		return NULL;
	}

	/* main loop */
	jcd->main_loop = g_main_loop_new (context, FALSE);
	g_main_loop_run (jcd->main_loop);

	return NULL;
}

/* start all jabber clients listed in the jabber_account_list */
static int jabber_client_start(void)
{
	struct jabber_account *jac = jabber_account_list;
	int res = 0;
	while (jac) {
		if (ast_pthread_create(&jac->client_thread, NULL, do_jabber_client, jac) < 0) {
			ast_log(LOG_ERROR, "Unable to start jabber client thread [%s]\n", jac->config_section);
			res = -1;
		}
		jac = jac->next;
	}
	return res;
}

/* stop all jabber clients listed in the jabber_account_list */
static int jabber_client_stop(void)
{
	struct jabber_account *jac = jabber_account_list;
	struct jabber_client_data *jcd;

	while (jac) {
		jcd = jac->client_data;
		if (jcd) {
			/* stop connection & main loop */
			lm_connection_close (jcd->connection, NULL);
			g_main_loop_quit (jcd->main_loop);
			
			/* free jcd */
			free(jcd);
		}	
		/* kill thread */
		if (jac->client_thread && (jac->client_thread != AST_PTHREADT_STOP)) {
			pthread_cancel(jac->client_thread);
			pthread_kill(jac->client_thread, SIGURG);
			pthread_join(jac->client_thread, NULL);
		}
		jac->client_thread = AST_PTHREADT_STOP;
		jac = jac->next;
	}
	return 0;
}

/****************************************************************************/
/*** JABBER INIT ***/
/****************************************************************************/
int jabber_load(void)
{
	read_config();
	
	/* start jabber clients */
	jabber_client_start();
	
	return RESULT_SUCCESS;
}
int jabber_unload(void)
{
	/* kill all client threads */
	jabber_client_stop();
	
	/* clear jabber account list */
	clear_jabber_account_list();

	return RESULT_SUCCESS;
}

int read_config(void)
{
	struct ast_config *cfg;
	struct ast_variable *v;
	struct jabber_account *jac;
	
	cfg = ast_config_load(config);

	/* We *must* have a config file otherwise stop immediately */
	if (!cfg) {
		ast_log(LOG_NOTICE, "Unable to load config %s\n", config);
		return -1;
	}
	
	/* read the [general] section */
	v = ast_variable_browse(cfg, "general");
	while (v) {
		if (!strcasecmp(v->name, "register")) {
			jac = create_new_jabber_account();
			strcpy(jac->config_section, v->value);
		} else {
			ast_log(LOG_WARNING, "unknown configuration item at line %d of %s\n", v->lineno, config);
		}

		v = v->next;
	}

	/* read registered sections */
	jac = jabber_account_list;
	while (jac) {
		v = ast_variable_browse(cfg, jac->config_section);
		while (v) {
			if (!strcasecmp(v->name, "user")) {
				ast_copy_string(jac->user, v->value, sizeof(jac->user));
			} else if (!strcasecmp(v->name, "host")) {
				ast_copy_string(jac->host, v->value, sizeof(jac->host));
			} else if (!strcasecmp(v->name, "port")) {
				ast_copy_string(jac->port, v->value, sizeof(jac->port));
			} else if (!strcasecmp(v->name, "pass")) {
				ast_copy_string(jac->pass, v->value, sizeof(jac->pass));
			} else if (!strcasecmp(v->name, "resource")) {
				ast_copy_string(jac->resource, v->value, sizeof(jac->resource));
			} else if (!strcasecmp(v->name, "use_ssl")) {
				jac->use_ssl = ast_true(v->value);
			} else if (!strcasecmp(v->name, "ssl_fingerpring")) {
				ast_copy_string(jac->ssl_fingerprint, v->value, sizeof(jac->ssl_fingerprint));
			}
			v = v->next;
		}
		jac = jac->next;
	}

	/* Release configuration from memory */
	ast_config_destroy(cfg);

	return 0;
}

/****************************************************************************/
/*** JABBER ACCOUNTS ***/
/****************************************************************************/

static struct jabber_account* create_new_jabber_account(void)
{
	struct jabber_account *jac;

	/* create new jabber_account */
	jac = malloc(sizeof(struct jabber_account));
	
	/* set default values */
	memset(jac, 0, sizeof(struct jabber_account));
	strcpy(jac->port, "5222");
	strcpy(jac->resource, "asterisk");
	jac->client_thread = AST_PTHREADT_NULL;
	
	/* add new jabber_account to jabber_account_list */
	jac->next = jabber_account_list;
	jabber_account_list = jac;

	return jac;
}

static int clear_jabber_account_list(void)
{
	struct jabber_account *tmp;
	
	while (jabber_account_list) {
		tmp = jabber_account_list;
		jabber_account_list = jabber_account_list->next;
		free(tmp);
	}
	
	return RESULT_SUCCESS;
}

static struct jabber_account *find_jabber_account(char *config_section)
{
	struct jabber_account *jac = NULL;
	struct jabber_account *tmp = jabber_account_list;
	while (tmp && !jac) {
		if (!strcmp(tmp->config_section, config_section))
			jac = tmp;
		tmp = tmp->next;
	}
	return jac;
}


/****************************************************************************/
/*** APPLICATIONS ***/
/****************************************************************************/

/* send jabber message */
static char jabbersend_app_desc[] =
"Usage: JabberSend(config_section/user@host,message)\n"
"       send jabber message to user@host using the account given by config_section\n";

static int jabbersend_app(struct ast_channel *chan, void *data)
{
	char param[1024];
	char *dialstring;
	char *message;
	struct localuser *u;
	int res = 0;
	
	LOCAL_USER_ADD(u);
	
	if (ast_strlen_zero(data)) {
		ast_log(LOG_WARNING, "syntax error");
		LOCAL_USER_REMOVE(u);
		return -1;
	}
	
	ast_copy_string(param, (char*)data, sizeof(param));
	
	dialstring = param;
	message = strchr(param, '|');
	if (!message) {
		ast_log(LOG_WARNING, "syntax error [%s].\n", param);
		return -1;
	}
	*message = '\0';
	message++;
	
	res = jabber_send_message(dialstring, message);
	
	LOCAL_USER_REMOVE(u);

	return res;
}



static char jabberreceive_app_desc[] =
"Usage: JabberReceive(config_section/user@host,message)\n"
"       receive message from user@host using the account given by config_section\n";

static int jabberreceive_app(struct ast_channel *chan, void *data)
{
	char param[1024];
	struct localuser *u;
	int res = 0;
	char *new_msg = NULL;
	
	int timeout = 3600; /*< one hour default timeout */
	
	char *dialstring = NULL;
	char *message = NULL;
	char *config_section = NULL;
	char *jid;
	struct jabber_account *jac;
	
	LOCAL_USER_ADD(u);
	
	/* check for zero param */
	if (ast_strlen_zero(data)) {
		ast_log(LOG_WARNING, "syntax error");
		LOCAL_USER_REMOVE(u);
		return -1;
	}
	
	
	/* seperate dialstring & message and send message */
	ast_copy_string(param, (char*)data, sizeof(param));
	dialstring = param;
	message = strchr(param, '|');
	if (message) {
		*message = '\0';
		message++;
		res = jabber_send_message(dialstring, message);
	}
	
	/* seperate config_section & jid */
	split_dialstring(dialstring, &config_section, &jid);
	
	jac = find_jabber_account(config_section);
	if (!jac) {
		ast_log(LOG_ERROR, "unknown config section [%s]\n", config_section);
		return -1;
	}
	
	/* set partner_jid and clear old message */
	ast_copy_string(jac->partner_jid, jid, sizeof(jac->partner_jid));
	jac->new_message[0] = '\0';
	
	/* answer channel */
	if (chan->_state != AST_STATE_UP)
		ast_answer(chan);

	/* wait for new message */
	for (; !new_msg && timeout > 0; ) {
		if (*jac->new_message)
		{
			new_msg = jac->new_message;
			break;
		}
		timeout--;
		if (chan->_state != AST_STATE_UP || ast_safe_sleep(chan, 1000))
			break;
	}
	jac->partner_jid[0] = '\0';
	
	/* store new msg in JABBER_MSG var */
	if (chan->_state == AST_STATE_UP) {
		if (new_msg)
			pbx_builtin_setvar_helper(chan, "JABBER_MSG", new_msg);
		else
		{
			/* probably timeout or hangup */
			pbx_builtin_setvar_helper(chan, "JABBER_MSG", "");
			res = -1;
		}
	}
	else
		res = -1;
	
	LOCAL_USER_REMOVE(u);

	return res;
}

/****************************************************************************/
/*** CLI ***/
/****************************************************************************/

static char cli_jabber_reload_usage[] =
"Usage: jabber reload\n"
"       reload jabber.\n";

static int cli_jabber_reload(int fd, int argc, char *argv[])
{
	int result = jabber_unload() & jabber_load();
	
	if (result == RESULT_SUCCESS) {
		ast_log(LOG_NOTICE, "jabber reloaded\n");
	} else {
		ast_log(LOG_ERROR, "failed to reload jabber\n");
	}

	return result == RESULT_SUCCESS ? result : RESULT_FAILURE;
}

static char cli_jabber_show_accounts_usage[] =
"Usage: jabber reload\n"
"       reload jabber.\n";

static int cli_jabber_show_accounts(int fd, int argc, char *argv[])
{
	struct jabber_account *jac = jabber_account_list;
	struct jabber_client_data *jcd;
	char online_status[256];
	LmSSL *ssl;
	const char *fingerprint;
	
	while (jac) {
		
		/* determine online status */
		if (!jac->client_data)
			ast_copy_string(online_status, "[uninitialized]", sizeof(online_status));
		else {
			jcd = jac->client_data;
			if (!jcd->connection)
				ast_copy_string(online_status, "[no connection]", sizeof(online_status));
			else if (!lm_connection_is_open(jcd->connection))
				ast_copy_string(online_status, "[closed connection]", sizeof(online_status));
			else {
				ssl = lm_connection_get_ssl(jcd->connection);
				if (!ssl)
					ast_copy_string(online_status, "[connected]", sizeof(online_status));
				else {
					fingerprint = lm_ssl_get_fingerprint(ssl);
					if (!fingerprint)
						ast_copy_string(online_status, "[connected with SSL]", sizeof(online_status));
					else
						snprintf(online_status, sizeof(online_status), "[connected with SSL {%s}]", fingerprint);
				}
			}
		}
		
		ast_cli(fd, "%s:%s@%s:%s/%s %s %s\n", jac->user, jac->pass, jac->host, jac->port, jac->resource,
				jac->use_ssl ? "(SSL)" : "", online_status);
		jac = jac->next;
	}
	
	return RESULT_SUCCESS;
}

static struct ast_cli_entry  my_clis[] = {
	{ { "jabber", "reload", NULL }, cli_jabber_reload, "reload module", cli_jabber_reload_usage },
	{ { "jabber", "show", "accounts", NULL }, cli_jabber_show_accounts, "show configured accounts", cli_jabber_show_accounts_usage },
};


/****************************************************************************/
/*** ASTERISK MODULE INTERFACE ***/
/****************************************************************************/

/*! \brief  load_module: PBX load module - initialization ---*/
static int _load_module(void)
{
	/* jabber init */
	jabber_load();
	
	/* CLI */
	ast_cli_register_multiple(my_clis, sizeof(my_clis)/ sizeof(my_clis[0]));

	/* register apps */
	ast_register_application("JabberSend", jabbersend_app, "send jabber messages", jabbersend_app_desc);
	ast_register_application("JabberReceive", jabberreceive_app, "receive jabber messages", jabberreceive_app_desc);

	return 0;
}
int load_module() { return _load_module(); }
static int _unload_module(void)
{
	/* unregister apps */
	ast_unregister_application("JabberSend");
	ast_unregister_application("JabberReceive");
	
	STANDARD_HANGUP_LOCALUSERS;

	/* CLI */
	ast_cli_unregister_multiple(my_clis, sizeof(my_clis) / sizeof(my_clis[0]));

	/* jabber shutdown */
	jabber_unload();

	return 0;
}
int unload_module() { return _unload_module(); }

/*! \brief  reload: Part of Asterisk module interface ---*/
int reload(void)
{
	return _unload_module() & _load_module();
}

int usecount()
{
	int res;
	STANDARD_USECOUNT(res);
	return res;
}

char *key()
{
	return ASTERISK_GPL_KEY;
}

char *description()
{
	return (char *) desc;
}

